import numpy as np
import pandas as pd
from imblearn.over_sampling import SMOTE
import seaborn as sns
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn import model_selection
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score, roc_auc_score
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn import svm
from sklearn.model_selection import GridSearchCV
from sklearn.feature_extraction import FeatureHasher

def generar_muestras_balanceo(dataset, columna):
    """ Esta función permite crear muestras sintéticas, a partir de conjuntos de datos de clasificación desequilibrados.
        Para emplear esta función es necesario (import : Numpy, pandas, from imblearn.over_sampling import SMOTE)
        Se ayuda de la función SMOTE (Tecnica de sobremuestreo de minorias sintéticas)
        Se basa en la creación de muestras sinteticas, en el grupo/grupos minoritario(s).
    Parameters:
            -dataset: Set de datos desbalanceado
            -columna: Tiene que estar incluída en el set de datos desbalanceado, puede ser de tipo "string" ó "númerico". 
                      Esta columna representara a la clase o grupo que pertence mi muestra (ejemplo: Tipo, Clase, Grupo, Ypredict)
    
    Output: Regresa el dataset balanceado (el cual contiene las muestras originales, como las muestras creadas)
    """
    X= dataset.drop([columna],1)
    y= dataset[columna]

    smt= SMOTE(random_state=42)
    X_train_sm, y_train_sm = smt.fit_resample(X, y)

    dataset_balanceado= pd.concat([X_train_sm,y_train_sm], axis=1)
    
    return dataset_balanceado


def balanceado(df, target):
    list_b = []

    for index,value in df[target].value_counts().items():
        list_b.append({'Categoria': index, 'Peso': round(float(value )/ len(df[target])*100, 2)})

    list_b = pd.DataFrame(list_b)

    return list_b
    

def correlation_coeff(dataset,threshold):
    ''' 
        The correlation coefficient is a statistical measure of the strength of the relationship 
        between the relative movements of two variables.
        Parameters X_train ,Threshold where threshold is 
        the maximimum percent in correlation in the independants variables will be high correlated.
        X_train is the dataset, in this case we just interestd on the indepedant features, that wy we use X_train.
        threshold values must bue from 0 to 1 like, 0.5, 0.7, 0.9.
        Function will return witch variable have a high correlation in indepeant features.
    '''
    col_corr = set() #Set of all the names of corralated columns
    corr_matrix = dataset.corr()
    for i in range(len(corr_matrix.columns)):
        for j in range(i):
            if corr_matrix.iloc[i, j] > threshold: # we are interested in absolute coeff value
                colname = corr_matrix.columns[i] #guetting the name of column
                col_corr.add(colname)
    return col_corr


def balance_observation(df, target):
    '''
    Function to know the balance classes.
    Input:
        - df: dataframe
        - target: column
    Output: 
        Dataframe with the categories of the column and their weight.  

    Libraries used:
        import pandas as pd
        import numpy as np    
    '''

    new_list = []
    for index, value in df[target].value_counts().items():
        new_list.append({'Category': index, 'Weight': round(float(value )/ len(df[target])*100, 2)})

    return pd.DataFrame(new_list)


def replace_missings_train_test(X_test, X_train, col):
    '''
	Function to eliminate or replace missings. This version is for dataframes that have been separated into train and test.
	The user is asked for the dataframe and the column that they want to act on, and they are given options to choose what they want to
	do with the missings.

	Parameters:
	                - df: dataframe
	                - col: column you want to modify

	Libraries used:
	    import pandas as pd
	    import numpy as pd                
	'''

    tipo = str(X_test[col].dtype)
    if 'int' in tipo or 'float' in tipo:
        op = str(input("Enter Mean to change your missings to the mean or Median to change them to the median"))
        if op.lower() == "mean":
            X_train[col] = X_train[col].fillna(X_train[col].mean())
            X_test[col]  = X_test[col].fillna(X_train[col].mean())
            print("The missings of your column have been replaced by the mean of train on train and test sets")
        elif op.lower() == "median":
            X_train[col] = X_train[col].fillna(X_train[col].median())
            X_test[col]  = X_test[col].fillna(X_train[col].median())
            print("The missings of your column have been replaced by the median of train on train and test sets")
    else:
        print("Your variable is not numerical, changing the missings in train and test sets by the mode of train...")
        X_train[col] = X_train[col].fillna(X_train[col].mode().iloc[0])
        X_test[col]  = X_test[col].fillna(X_train[col].mode().iloc[0])
        print("The missings of your column have been replaced in train and test sets by the mode of train")


def models():
    """
    Function which summarizes the different machine learning models available for supervised learning
    Inputs:
    
    Output:
        Description of the different models suitable for classifcation or regression problems.
    """


    text= """Below you will find a list with the available machine learning models for supervised learning included in this library.

    According to the desired results of your model (regression or classification) you may choose among the following:

        Classification

            - Logitic regression: Logistic regression is a statistical analysis method to predict a binary outcome, such as yes or no, based on prior observations of a data set.
              
              A logistic regression model predicts a dependent data variable by analyzing the relationship between one or more existing independent variables.
            
            - SVM: Or Suport Vector Machine,  the objective of the support vector machine algorithm is to find a hyperplane in an N-dimensional space(N — the number of features) that distinctly classifies the data points.

            - Decision tree: Decision trees are an approach used in supervised machine learning, a technique which uses labelled input and output datasets to train models.
              
              The approach is used mainly to solve classification problems, which is the use of a model to categorise or classify an object.

        Regression

            - Linear regression: Linear Regression is a supervised machine learning algorithm where the predicted output is continuous and has a constant slope.
              
              It's used to predict values within a continuous range, (e.g. sales, price) rather than trying to classify them into categories (e.g. cat, dog).

        Classification & Regression

            - KNN: The abbreviation KNN stands for “K-Nearest Neighbour”. It is a supervised machine learning algorithm. The algorithm can be used to solve both classification and regression problem statements.
              
              The number of nearest neighbours to a new unknown variable that has to be predicted or classified is denoted by the symbol 'K'.

            - Random Forest: Random Forest is a robust machine learning algorithm that can be used for a variety of tasks including regression and classification.
              
              It is an ensemble method, meaning that a random forest model is made up of a large number of small decision trees, called estimators, which each produce their own predictions.
              
              The random forest model combines the predictions of the estimators to produce a more accurate prediction. """
    
    return print(text)


def best_param_clf(X_train, y_train, model):
    """This function allows you to select the best parameters for your machine learning model.
    
    You may choose the model among the following: 'Logistic' = LogisticRegression, 'Kneighbors' = KNeighborsClassifier, 'RandomForest' = RandomForestClassifier, 'SupportVC' = SVC, 'DecisionTree' = DecisionTreeClassifier

    The following libraries should be implemented
    from sklearn.linear_model import LogisticRegression
    from sklearn.tree import DecisionTreeClassifier
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.svm import SVC
    from sklearn import svm
    from sklearn.model_selection import GridSearchCV
"""

    if model == "Logistic":
        clf = LogisticRegression()
        parameter_grid = {'C': [0.01, 0.1, 1, 2, 10, 100], 'penalty': ['l1', 'l2'], 'solver' : ['liblinear','lbfgs']}
        gridsearch = GridSearchCV(clf, parameter_grid)
        gridsearch.fit(X_train, y_train)
        return print("The best parameters for the Logistic Regression model are: ", gridsearch.best_params_)
    elif model == "KNeighbors":
        clf = KNeighborsClassifier()
        parameter_grid = {"n_neighbors": [3,5,7,9,11],"weights": ["uniform", "distance"]}
        gridsearch = GridSearchCV(clf, parameter_grid)
        gridsearch.fit(X_train, y_train)
        return print("The best parameters for the K Neighbors Classifier model are: ", gridsearch.best_params_)
    elif model == "RandomForest":
        clf =  RandomForestClassifier()
        parameter_grid = {"n_estimators": [100,120,150,200,300,500],"max_depth": [3,4,5,6,10,15,17],"max_features": ["sqrt", 3, 4]}
        gridsearch = GridSearchCV(clf, parameter_grid)
        gridsearch.fit(X_train, y_train)
        return print("The best parameters for the Random Forest Classifier model are: ", gridsearch.best_params_)
    elif model == "SuportVC":  
        clf = SVC()
        parameter_grid = {"C": [0.01, 0.1, 0.3, 0.5, 1.0, 3, 5.0, 15, 30], "kernel": ["linear", "poly", "rbf"], "degree": [2, 3, 4, 5],"gamma": [0.001, 0.1, "auto", 1.0, 10.0, 30.0]}
        gridsearch = GridSearchCV(clf, parameter_grid)
        gridsearch.fit(X_train, y_train)
        return print("The best parameters for the SVC model are: ", gridsearch.best_params_)
    elif model == "DecisionTree":
        clf = DecisionTreeClassifier()
        parameter_grid = {"max_depth": list(range(1,10))}
        gridsearch = GridSearchCV(clf, parameter_grid)
        gridsearch.fit(X_train, y_train)
        return print("The best parameters for the Decision Tree Classifier model are: ", gridsearch.best_params_)


def classif_metrics(pred, y_test):
    """ 
    This function allows you to have the metrics of your classification model
    
    Syntax: pred(Prediction X_test(y_pred)), y_test
    
    "pred" has to be defined "as is" outside the function
    
    Required Libraries:
    from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score, roc_auc_score, roc_curve, precision_recall_curve, confusion_matrix
    """
    metrics = ["Accuracy", (accuracy_score(pred, y_test))], ["Precision", (precision_score(pred, y_test))],["Recall", (recall_score(pred, y_test))],["F1", (f1_score(pred, y_test))], ["AUC", (roc_auc_score(pred, y_test))] 
    metrics_df = pd.DataFrame(metrics, columns = ['METRICS', 'Results'])
    metrics_df.set_index("METRICS", inplace= True)
    return print(metrics_df)


def hashing(X_train, name_column, new_clm_1, new_clm_2, new_clm_3, new_clm_4, new_clm_5):
    ''' Con está función se puede tratar una característica categórica de alta cardinalidad para obtener varias columnas númericas sin perder información
    
    Required Libraries:
    from sklearn.feature_extraction import FeatureHasher
    '''

    h = FeatureHasher(n_features=5, input_type='string')
    f = h.transform(X_train[name_column])
    f.toarray()

    hash_df= pd.DataFrame(f.toarray())

    hash_df = hash_df.rename(columns={0:new_clm_1,
                        1:new_clm_2,
                        2:new_clm_3,
                        3:new_clm_4,
                        4:new_clm_5})

    hash_df.set_index(X_train.index,inplace = True)

    X_train = pd.concat([hash_df, X_train],axis = 1 )
    X_train= X_train.drop([name_column], axis=1)
    return X_train


#from .Cleaning import remove_all_characters, filter_by_value, decoding, detect_outlier, missings, replace_missings, change_type, modify_datetime, modify_characters, remove_all_characters
#from .Visualization import grid_plot, plot_predictions, correlation_test, c_mat, data_report_numeric, data_report_categ, data_report, countplot, pie_random, relplot, plot_feature_importance, bar_line_plot
#from .Learning import generar_muestras_balanceo, balanceado, correlation_coeff, replace_missings_train_test, models, best_param_clf, classif_metrics, hashing


import matplotlib.pyplot as plt
import seaborn as sns
from math import ceil
import pandas as pd
import numpy as np
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
from scipy.stats import skew
import random

def grid_plot(df, y, hue=None, title_num= None, title_cat=None, 
            title_size= 11, axes_fontsize=9, fontweight= 'bold', 
            color_obj='MediumSeaGreen',color_num= 'MediumOrchid', save=None):

    '''
    The objetive of this function is to show the relationship between each of the attributes of X with respect to target.
    It also allows to include 'hue' to group by categories.

    In the case of non-categorical variables it generates scatterplots and if they are categorical it generates histograms.

    The function allows you to enter a title for categorical graphs and another for non-categorical graphs.

    You can also change the font size and width, the colors of the graphs and decide whether to save them.

    To run the function correctly it is mandatory to import the following libraries: seaborn, matplotlib.pyplot, pandas and 
    from math import ceil.  

    It returns each of the graphs in grid format.
    '''

    fig = plt.figure(figsize=(15,15))
    fig.subplots_adjust(hspace=0.4, wspace=0.4)

    if ceil(len(df.columns)/3) < 1:
        mnrows = 1
    else:
        mnrows = ceil(len(df.columns)/3)

    if mnrows < 3:
        mncols=mnrows
    else:
        mncols = 3

    for i, column in enumerate(df.columns):
        ax = fig.add_subplot(mnrows, mncols, i + 1)
        if df.dtypes[column] != object:

            sns.scatterplot(x = df.iloc[:,i], y = y, ax= ax, hue= hue, color=color_num)
            ax.set_title(title_num, fontsize=title_size)
            ax.set_xlabel(column, fontsize=axes_fontsize, fontweight= fontweight)

        else:
            df[column].value_counts().plot(kind="hist", ax= ax, color=color_obj)
            ax.set_title(title_cat, fontsize=title_size, fontweight= fontweight)
            ax.set_xlabel(column, fontsize=axes_fontsize, fontweight= fontweight)

    plt.show();
    
    if save != None:
        text= input('Enter the name you want to save the image: ')
        plt.savefig('./{}.png'.format(text))


def plot_predictions(ind_x, y_test, pred, title_x= '',title_y= '', new_prediction_x=None, new_prediction=None, save=None):

    '''
    The objective of this function is to plot in a graph the values of y_test with respect to the predictions made with the ML model.

    It is mandatory to put the index of X, y_test and the predictions.

    It is allowed to modify the title of the axes, enter a new prediction and save the plot.

    To enter the new prediction, it is necessary to input the X and y value of the new prediction.

    To run the function correctly it is mandatory to import the following libraries: matplotlib.pyplot and pandas.

    This function returns a dot plot with the entered values.
    '''

    plt.figure(figsize=(20, 7))
        
    p_df = pd.DataFrame()
        
    p_df['indiceX'] = ind_x
    p_df['real'] = y_test
    p_df['prediccion'] = pred

    x = p_df['indiceX']
    y1 = p_df['real']
    y2 = p_df['prediccion']

    plt.plot(x, y1,'gp',label='Actual', alpha=0.8)
    plt.plot(x, y2, 'mp', label='Prediction', alpha=0.5)
    
    if new_prediction != None:
        plt.plot(new_prediction_x, new_prediction, 'ko', label= 'New prediction')
        
    plt.title("Actual vs Prediction",fontsize=13)
    plt.xlabel(title_x,fontsize=11)
    plt.ylabel(title_y,fontsize=11)
    plt.xticks(rotation=70)
    plt.legend()

    plt.show();

    if save != None:
        text= input('Enter the name you want to save the image: ')
        plt.savefig('./{}.png'.format(text))


def correlation_test(df, color_phik= 'PiYG', color_p= 'coolwarm', save=None):

    '''
    The objective of this function is to have in a single code the heatmaps of the Pearson and Phik coefficients.

    It has as mandatory parameter a dataframe.
    The optional parameters are: change the color of any of the heatmaps and the option to save them.
    
    To run the function correctly it is mandatory to import the following libraries: seaborn, matplotlib.pyplot \
     and from phik import phik_matrix.
    
    This function returns each of the heatmaps correctly identified with the title.
    '''

    fig, axes = plt.subplots(2, 1, figsize=(15,10))
    plt.xticks(rotation = 90)
    sns.heatmap(df.phik_matrix(), cmap= color_phik, annot=True, ax= axes[0], linewidths=.5)
    axes[0].set_title(label= 'Correlation Phik', fontsize=14, fontweight= 'bold')
    
    sns.heatmap(df.corr(), cmap= color_p, annot=True, ax=axes[1], linewidths=.5)
    axes[1].set_title(label= 'Correlation Pearson', fontsize=14, fontweight= 'bold')
    
    plt.show();
    
    if save != None:
        text= input('Enter the name you want to save the image: ')
        plt.savefig('./{}.png'.format(text))


def c_mat(y_test, X_test, modelo, xticklabels=None, yticklabels=None, save= None):
    '''
    The y_test parameter must contain the y_test of the model being tested, just as the X_test must contain the X_test of the model being tested.

    The model parameter is the model for which we will represent the confusion matrix.

    xticklabels needs a list of labels to assign to the x-axis values. The default value is None.

    yticklabels needs a list of labels to assign to the y-axis values. The default value is None.
    
    The default save parameter is None, which reflects whether the graph will be saved or not.

    '''

    predictions = modelo.predict(X_test)

    c_mat = confusion_matrix(y_test, predictions)

    sns.heatmap(c_mat/c_mat.sum(axis=1),
                annot=True,
                cmap='Spectral',
                cbar=True,
                linewidths=3,
                linecolor='w',
                square=True,
                xticklabels=xticklabels,
                yticklabels=yticklabels)
    plt.show();
    
    print(classification_report(y_test, predictions))

    if save != None:
        text= input('Introduce el nombre con el que deseas guardar la imagen: ')
        plt.savefig('./{}.png'.format(text))


def data_report_numeric(df):
    '''
    Function to make a summary table of the numerical variables of the dataset.
    parameters:
            -df:dataframe
    Output:
            Dataframe with the following data for each column:
                  n_instances, mean, std, min, max, q25, q50, q75, missings (%), outliers (%) and skew
    Libraries:
            - import numpy as np
            - from scipy.stats import skew
    '''
    # to know which columns are numeric
    list_columns = df.select_dtypes(include=np.number).columns

    dict_num = {}    

    for columna in df[list_columns]:

        count = len(df[columna])
        percent_missing = round(df[columna].isnull().sum()*100 / len(df[columna]), 2)

        q25 = np.percentile(df[columna], 25)
        q50 = np.percentile(df[columna], 50)
        q75 = np.percentile(df[columna], 75)
        iqr = q75 -q25

        cutoff = iqr * 1.5
        lower = q25 - cutoff
        upper = q75 + cutoff
        
        # n outliers
        n_outliers = 0
        for x in df[columna]:
            if x < lower or x > upper:
                n_outliers += 1
            percent_outliers = round(n_outliers / count *100, 2)  

        mean = df[columna].mean()
        maximo = df[columna].max()
        minimo = df[columna].min()
        desv = df[columna].std()

        skew_column = skew(df[columna])

        dict_num[columna] = [count, mean, desv, minimo, maximo, q25, q50, q75, percent_missing, percent_outliers, skew_column]

    report_num_df = pd.DataFrame(dict_num, )
    report_num_df.index = ['n_instancess', 'mean', 'std', 'min', 'max', 'q25', 'q50', 'q75', 'missings (%)', 'outliers (%)', 'skew']

    return report_num_df


def data_report_categ(df):
    '''
    Function to make a summary table of the categorical variables of the dataset.
    parameters:
            -df:dataframe
    Output:
            Dataframe with the following data for each column:
                - number of observations
                - number of different categories
                - % cardinality
                - fashion
                - weight (%) of the mode
    Libraries:
            - import numpy as np
            - import pandas as pd               
    '''
    
    dict_cat = dict()

    for column in df:
        #if column not numeric
        if column not in df.select_dtypes(include=np.number).columns:

            n_observations = len(df[column])
            n_cat = df[column].nunique()
            perc_cardin = round(n_cat/n_observations * 100, 2)
            mode = df[column].mode()[0]
            w_mode = round(len(df[df[column] == df[column].mode()[0]]) / df.shape[0] * 100, 2)

            dict_cat[column] = [n_observations, n_cat, perc_cardin, mode, w_mode]

        table = pd.DataFrame(dict_cat)
        table.index = ['n_instances', 'n_categories', 'cardin (%)', 'mode', 'mode (%)']

    return table


def data_report(df):
    '''
    General data report.
    Parameters:
                - df: dataframe
    Output:
            Dataframe con:
                - Column name
                - Data type
                - % Missings 
                - % Cardinalidad 
    
    Libraries used:
        import pandas as pd
        import numpy as pd
    '''
    
    # COLUMN NAMES
    cols = pd.DataFrame(df.columns.values, columns = ['COL_N'])

    # COLUMNS TYPES
    types = pd.DataFrame(df.dtypes.values, columns = ['DATA_TYPE'])

    # MISSINGS --> df % nans per columns
    percent_missing = round(df.isnull().sum()*100 / len(df), 2)
    percent_missing_df = pd.DataFrame(percent_missing.values, columns = ['MISSINGS (%)'])
    
    # UNIQUE VALUES --> % unique values per column
    unicos = pd.DataFrame(df.nunique().values, columns = ['UNIQUE_VALUES'])

    percent_candin = round(unicos['UNIQUE_VALUES']*100 / len(df), 2)
    percent_candin_df = pd.DataFrame(percent_candin.values, columns = ['CARDIN (%)'])

    # Unimos
    concatenado = pd.concat([cols, types, percent_missing_df, unicos, percent_candin_df], axis = 1)
    concatenado.set_index('COL_N', drop = True, inplace = True)

    return concatenado.T


def countplot(count,hue,df, title=None, save=None): 
    '''
            + In the first parameter we have to introduce the column we want to make the visualization with.

            + For the second parameter we have to introduce the column with which we want to fill these data.

            + The third parameter needs the original dataframe where these columns are.

            + For the fourth parameter we have to indroduce the title to the graph. This is optional, since by default it is None.

            + The fifth parameter gives the option to save the image with the name you prefer, default is None;
              if we want to save the image we have to set save=True, it returns a message confirming that the image has been saved.

            + You can also use the parameters without putting the initial declaration, but if you put one, you have to enter all of them.
              if you don't introduce the declaration, you have to follow the order: count, hue, df, savepic.
            
            --> Library requirements: pandas, matplotlib.pyplot as plt, and seaborn as sns.
        
    '''

    plt.figure(figsize=(15,5));
    sns.countplot(x=count, hue=hue, palette='BuPu', data=df)
    plt.title(title,  fontsize=20)
    plt.ylabel(' ')
    plt.xlabel(' ')
    plt.tight_layout();
    plt.show()
    
    if save != None:
        text = input("Enter the name you want to save the image with: ")
        plt.savefig('./{}.png'.format(text))
        print("Image saved under the name: {}".format(text))


def pie_random(df, title=None, save=None):
    '''
            + In the first parameter we have to introduce the data that we want to visualize, 
              this already makes a value_counts(), ej: df = df.

            + The second parameter is for the title, by default it is None, if we want to enter it we have to put it as follows: title= 'here the title'.

            + The third parameter gives the option to save the image with a random number, default is None;
              if we want to save the image we have to set save=True, it returns a message confirming that the image has been saved.

            + You can also use the parameters without putting the initial declaration, but if you put one, you have to enter all of them.
              If you do not put the declaration, you have to follow the order: df, title, savepic.

            --> Library requirements: pandas, numpy, random, matplotlib.pyplot.
            --> Recommendations: if you need to introduce more variables, we recommend a countplot from the same library.
    '''
    
    df = df.value_counts()
    colors = ["#90EE90", "#66B3FF", "#D87093", "#FFB6C1", "#F08080", "#FFCC99" , "#003f5c", "#7982B9", "#58508d", "#FF9999", "#74BBFB", "#9370D8", "#99FF99", "#BA55D3"]
    fig1, ax1 = plt.subplots()
    ax1.pie(df,colors=colors, labels=df.index, autopct="%1.1f%%", startangle=90)
    centre_circle = plt.Circle((0,0),0.80,fc="white")
    fig = plt.gcf()
    fig.gca().add_artist(centre_circle)
    ax1.axis("equal")
    plt.title(title)
    fig.show()
    plt.tight_layout();

    # images are saved in random mode
    if save == True:
        plt.savefig('./{}.png'.format(random.randint(0, 1000)))
        print("Savefig ok")


def relplot(df, values, hue, save=None):
    sns.relplot(data=df, 
    x=values, y=hue, 
    height=5, aspect=2, 
    kind="line")
    plt.ylabel('')
    if save != None:   
        text = input("Enter the name you want to save the image under: ")
        plt.savefig('./{}.png'.format(text))
        print("Image saved with the name of: {}".format(text))


def plot_feature_importance(importance, X, model_type, save=None):

    '''
    The importance parameter is the model from which the features importances will be output.

    The parameter X must be the dataframe with all the variables of the model with the names of the columns (without the target)
    
    The default save parameter is None, this parameter reflects whether the graph will be saved or not.
    '''

    importance = importance.feature_importances_

    feature_importance = np.array(importance)


    names=X.columns

    feature_names = np.array(names)

    data={'feature_names':feature_names,'feature_importance':feature_importance}
    fi_df = pd.DataFrame(data)

    fi_df.sort_values(by=['feature_importance'], ascending=False,inplace=True)

    plt.figure(figsize=(10,8))
    sns.barplot(x=fi_df['feature_importance'], y=fi_df['feature_names']) # palette= ? o dejo el style?
    
    plt.title(model_type + ' Feature Importance')
    plt.xlabel('Feature Importance')
    plt.ylabel('Featura Names')
    
    plt.show();
    
    if save != None:
        text = input("Enter the name you want to save the image with: ")
        plt.savefig('./{}.png'.format(text))
        print("Image saved under the name: {}".format(text))


def bar_line_plot(df, columna_bar, columna_line=None, title=None, label_bar=None, label_line=None, label_x=None, save=None):

    '''
    df is the DataFrame being used.

    column_bar is the name of the column you want to be represented by the bars.

    column_line is the name of the column you want to be represented with the line.

    The following parameters are set as None by default, therefore, you can choose wheater to write something or not:
    
        title is the title of the chart
    
        label_bar is the title of the bars
    
        label_line is the title of the line
    
        label_x is the title of the x-axis, the index of your dataframe

    Requirements: it is necessary to introduce a dataframe with index for the rows.
    '''

    fig = plt.figure()
    

    plt.title(title)

    ax = df[columna_bar].plot(kind='bar',
                                grid=True,
                                color='#2b3f48')

    ax.set_ylabel(label_bar)
    ax.set_xlabel(label_x)


    if columna_line!=None:

        ax1 = ax.twinx()
        ax1.plot(df[columna_line].values,
                linestyle='-',
                linewidth=2.0,
                color='#f98c8d',
                marker='o')
        

        ax1.set_ylabel(label_line)
        
    plt.show();


    if save != None:
        text= input('Introduce el nombre con el que deseas guardar la imagen: ')
        plt.savefig('./{}.png'.format(text))


import pandas as pd
import numpy as np

def remove_all_characters(df, filter, filter2 = None, filter3 = None, filter4 = None):
   
    """
    Function to remove all special characters in Dataframe. You can choose a specific character/wordl or remove all special characters. 
    Syntax:
        df = Dataframe 
        Filter = a character o list of character to remove in dataframe, you can use "all" for remove all character special
        You can use list with diferents things to remove in dataframe
        Filter2,  Filter3 and filter 4 is and aditional filters

    Libraries used:
        import pandas as pd
        import numpy as pd
        import re    
    """
    if filter == "all":
        df = df.replace(r"[\s\-\/;'`^¨_-]", "", regex=True)
    elif filter2 != None:
        df = df.replace(filter2, "", regex=True)
    elif filter3 != None:
        df = df.replace(filter3, "", regex=True)
    elif filter4 != None:
        df = df.replace(filter4, "", regex=True)
    else:
        df = df.replace(filter, "", regex=True)
    return df


def filter_by_value(df, filter_column, element, inverse = False, option=1, columns_select = None):
    '''
    To filter a dataframe by an specific element. 
    Input: 
        - df: dataframe
        - filter_column
        - element: filter element
        - option: to choose output, default: 1.
        - columnas_select: if opcion = 2, enter a list with the columns you want to get.
                           ejemplo: [col1, col2, col3]
        - filtro_inverso: default False --> to select the rows where the filter element appears.
                          True --> to select the rows where the filter element doesn't appear.
    Output:
        If opcion=1 --> complete dataframe filtered by the given element and column
        If opcion=2 --> dataframe with the columns in 'columnas select', filtered by the element and column given.   
    
    Libraries used:
        import pandas as pd
        import numpy as pd
    '''
    
    df_filtro = df[df[filter_column] == element]
    df_filtro_inverso = df[~(df[filter_column] == element)]
    
    if inverse == False:

        # Opcion 1 --> devuelve el df solo con las filas donde aparezca el element dado
        if option == 1:
            return df_filtro

        # Opcion 2 --> devuelve las columnas seleccionadas con las filas donde la columna filtro == element
        elif option == 2:
            return df_filtro[columns_select]

    elif inverse == True:

        # Opcion 1 --> devuelve el df filtrado 
        if option == 1:
            return df_filtro_inverso

        # Opcion 2 --> devuelve las columnas seleccionadas filtradas
        elif option == 2:
            return df_filtro_inverso[columns_select]
    else:
        print('''El parámetro inverso solo admite los valores True o False.
                False, por defecto, devuelve todas las columnas del dataframe.
                True, devuelve solo las columnas dadas''') 


def decoding(url):

  encodings=['utf-8','utf8', 'latin-1', 'latin1','iso-8859-1', 'iso8859-1', 'mbcs' , 'ascii', 'us-ascii', 'utf-16', 'utf16', 'utf-32', 'utf32']

  for i in encodings:
    try:
      df = pd.read_csv(url, encoding=i)
      return (df,i)

    except:pass


def detect_outlier(df,column):
    """
    You can use this function (IQR), several quartile values,
    and an adjustment factor to calculate
    boundaries for what constitutes minor and major outliers
    """
    sort_data = df[column].sort_values()
    quantile_1,quantile_3 = np.percentile(sort_data,[25,75])
    iqr_value = quantile_3 - quantile_1
    print('iqr value : ',iqr_value)
    
    ## Find the lower bound value and the  higher bound value
    lower_bound = quantile_1 - (1.5 * iqr_value)
    upper_bound = quantile_3 + (1.5 * iqr_value)

    #printing upper bound and lower_bound
    print('lower vale :',lower_bound,', higher value :',upper_bound)
    
    # Return all the data value that are outlier
    global outliers 
    outliers =  df[(df[column] > upper_bound) | (df[column] < lower_bound)][column]

    # % outliers
    print("Outliers (%):", round(len(outliers) / len(df[column]) * 100, 2))

    print("This are your outliers")
    print(outliers)
    print("Your outliers are stored in the variable : outliers, if you want to remove them you can use : df = df.drop(df.YOUR_COLUMN_NAME[outliers], axis = 0)")


def missings(df):
    '''
    Function that checks the missings 
    of the entire dataset and return the quantity and the percentage that are in each column
    parameters:
            -df:dataframe
    Libraries:
            - import pandas as pd
            - import numpy as pd       
    '''

    if df.isnull().sum().sum() > 0:  # the first sum calculates how many there are per column, and the second is the total of the dataset
        miss_total = df.isnull().sum().sort_values(ascending=False) 
        total = miss_total[miss_total > 0]

        miss_percent = df.isnull().mean().sort_values(ascending=False)
        percent = miss_percent[miss_percent > 0] *100

        missing_data = pd.concat([total, percent], axis=1, keys=['Total', 'Percent %'])

        print(f'Total and Percentage of Missings:\n {missing_data}')
    else: 
        print('No missings found.')


def replace_missings(df, col):
    '''
    Function to eliminate or replace missings. This version is for dataframes that have not been separated into train and test.
    The user is asked for the dataframe and the column that they want to act on, and they are given options to choose what they want to
    do with the missings.

    Parameters:
        - df: dataframe
        - col: column you want to modify
        
    Libraries used:
        import pandas as pd
        import numpy as pd               
    '''

    print("Your column has :", df[col].isnull().sum(), "missings, and is of type: ", df[col].dtype)
    m = df[col].isnull().sum()
    if m > 0:
        select = input("Enter Delete to remove your missings or Change to see the change options")
        if select.lower() == "delete":
            select2 = input("Enter rows to delete all your rows that contains missings, enter column to delete the column that contains missings")
            if select2.lower() == "rows":
                df = df.dropna(axis = 0, inplace = True)
                print("Your rows with missings have been removed")
            if select2.lower() == "column":
                df = df.drop(col, axis = 1, inplace = True)
                print("Your columns with missings have been removed")
    
        elif select.lower() == "change":
            tipo = str(df[col].dtype)
            if 'int' in tipo or 'float' in tipo:
                op = str(input("Enter Mean to change your missings to the mean or Median to change them to the median"))
                if op.lower() == "mean":
                    df[col] = df[col].fillna(df[col].mean())
                    print("The missings of your column have been replaced by the mean")
                elif op.lower() == "median":
                    df[col] = df[col].fillna(df[col].median())
                    print("The missings of your column have been replaced by the median")
            else:
                print("Your variable is not numerical, changing the missings by the mode...")
                df[col] = df[col].fillna(df[col].mode().iloc[0])
                print("The missings of your column have been replaced by the mode")
    else:
        print("There are no missings to deal with")


def change_type(df, columns, type):
    '''
    To change a column/s data types.
    Parameters:
            - df: pd.DataFrame
            - columns: list with the columns we want to transform [col1, col2, col3]
            - type:
                    str -> for string/object
                    int -> for integer
                    float -> for decimal
                    bool -> for boolean 
    Output:
            - Transforms the type of the given columns
            - Prints to the screen the final data type of the transformed columns

    Libraries used:
        import pandas as pd
        import numpy as pd        
    '''
    # cambia el tipo de las columnas
    for column in df[columns]:
        df[column] = df[column].astype(type, errors = 'raise') # podríamos cambiarlo a 'ignore'
    
    # devuelve impresión con los nuevos tipos de cada columna    
    print(df[columns].dtypes)


def modify_datetime(df, column, formato="%Y/%m/%d", to = "date", yearordey = None, change_format = None):
    """
    function to convert columns in datetimes or strings and modify format
    Syntax:
        df = dataframe
        column = column you want modify
        formato = format datetime
        to = if you want convert str to datetime or datetime to str
        yearorday = Parameter that allows to show only the year, the month or the day. It's optional
        change_format = you can change datetime format. It's opcional
    Usage:
        to -> can transform dtype of column, Datetime to str or str to datetime (to = date or str)
        yearorday -> can show only the year, month or day
        change_format -> can change format, example from 2001-02-08 to 08/02001
    Librerias :
    pandas, datetime.datetime
    """
    #Convert dtypes
    if (to == "date") and  (yearordey == None) and (change_format == None):
        df[column] = pd.to_datetime(df[column], format=formato)
        return df[column]
    elif (to == "str") and (yearordey == None) and (change_format == None):
        df[column] = df[column].apply(lambda x: x.strftime(formato))
        return df[column]

    
    #Convert format day year or month
    elif (yearordey == "y") and (df[column].dtype == "object") and (change_format == None):
        df[column] = pd.to_datetime(df[column], format=formato)
        df[column] = df[column].dt.year 
        return df[column]
    elif (yearordey == "m") and (df[column].dtype == "object") and (change_format == None):
        df[column] = pd.to_datetime(df[column], format=formato)
        df[column] = df[column].dt.month
        return df[column]
    elif (yearordey == "d") and (df[column].dtype == "object") and (change_format == None):
        df[column] = pd.to_datetime(df[column], format=formato)
        df[column] = df[column].dt.day 
        return df[column]
    elif (yearordey == "y") and (change_format == None):
        df[column] = df[column].dt.year 
        return df[column]
    elif (yearordey == "m") and (change_format == None):
        df[column] = df[column].dt.month
        return df[column]
    elif (yearordey == "d") and (change_format == None):
        df[column] = df[column].dt.day 
        return df[column]

    #Convert full format to other format

    elif (change_format != None) and  (df[column].dtype == "object"):
        df[column] = pd.to_datetime(df[column], format=formato)
        df[column] =  df[column].apply(lambda x: x.strftime(change_format))
        return df[column]
    elif change_format != None:
        df[column] =  df[column].apply(lambda x: x.strftime(change_format))
        return df[column]


def modify_characters(df, column, remove, add=" ", series = False):
    """Function to remplace a characters by a space or another character OR return a list with all modified elements.
        Syntax:
                df = dataframe
                column = column you want 
                remove = the character you want to delete, you can attach a list
 
                add = the caracter you want to remplace. (space for default)
                series = If true, it returns ONLY a list with all modified elements . (False for default)
                """
    if (series == False) and (type(remove) != int):
        df[column] = df[column].apply(lambda x: x.replace(remove, add))
        return df
    elif (type(remove) == int or float) and (series == False):
        df[column] = df[column].replace(remove, add)
        return df
    elif (series == True) and (type(remove) == int or float) and (series == False):
        modified_column_list = df[column].replace(remove, add)
        return modified_column_list
    elif (series == False) and (type(remove) != int):
        modified_column_list = df[column].apply(lambda x: x.replace(remove, add))
        return modified_column_list


def remove_all_characters(df, filter):
   
    """
    Function to remove all characters in Dataframe. You can choose a specific character or remove all special characters. 
    Syntax:
        df = Dataframe 
        Filter = a character o list of character to remove in dataframe, you can use "all" for remove all  special character
        You can use list with diferents things to remove in dataframe
        

    Libraries used:
        import re    
    """
    if filter == "all":
        df = df.replace(r"[\s\-\/;'`^¨_-]", "", regex=True)
    else:
        df = df.replace(filter, "", regex=True)
    return df